-- 3D Map test file, objective is to make a 10x10 mesh grid with height variation
-- Goal 1: Render 10x10 textured tiles with 3/4 perspective
-- -> Tiles are a set of four inward pointing triangles forming a square
-- -> Tiles can be rotated in 3D space and freely distorted once created
-- -> All tiles share a single draw call and texture map
-- -> Multi-mesh rendering may be needed for multiple textures, might skip this entirely
-- Goal 2: Once G1 complete, tiles will automatically generate additional 
--		   triangles based on Z variation between neighbors
-- -> Wall mapping could be tricky, base texture on origin or by an alg
-- -> Implement meta-textured single mesh rendering (I might do this earlier)
-- -> Saving of maps/meshes should be implemented by this point
-- Goal 3: Shaders and lighting
-- -> Add uh... shaders and lighting! Vertex lighting should be dead simple, others... not so much


globalOrigin = 0
globalRotation = 0
globalZoom = 0
started = true

-- NOTE ON DATA FORMATS
-- Some things use formatted data, which is defined below for manual object creation purposes
--[[ For convenience, this comment section is in blocks

vertex2D = { x,y,u,v,r,g,b,a } --  2D and 3D vertex coordinates with texture mapping and RGBA color data
vertex3D = { x,y,z,u,v,r,g,b,a } --Should be referred to with direct references, ie: vector[1] for x

quad = { x,y,w,h } -- Basic 2D box shape, not often used

tri2D = { vertex2D, vertex2D, vertex2D } -- Triangle shape made of 3 vertices, can be 2D or 3D
tri3D = { vertex3D, vertex3D, vertex3D }

object2D = { tri2D, tri2D, tri2D, ... } -- Objects finalized and ready for rendering or pre-render are triangle sets
object3D = { tri3D, tri3D, tri3D, ... } -- Allows for easy clockwise checking

basicBox = { topLeft,bottomLeft,center,topRight,bottomRight } -- Each entry is a vertex2D or vertex3D
		   -- Correct order for five point box shape for correct triangle calculation
		   -- Noted here for manual test object creation or for function reference

vertexDrawSet = { vertex2D, vertex2D, vertex2D, ... } -- vertex draw sets should be single-level vertex2D tables
				-- final result of any mesh merging should result in one of these per draw call and no more

--]]

function newVertex2D(x,y,u,v,r,g,b,a)
	-- instantiates a new vertex2D with some default values
	-- vertex = newVertex2D() produces vertex = {0,0,0,0,255,255,255,255}
	-- function exists mostly as shorthand :D
	local x = x or 0
	local y = y or 0
	local u = u or 0
	local v = v or 0
	local r = r or 255
	local g = g or 255
	local b = b or 255
	local a = a or 255
	return {x,y,u,v,r,g,b,a}
end

function newVertex3D(x,y,z,u,v,r,g,b,a)
	-- instantiates a new vertex3D with some default values
	-- vertex = newVertex3D() produces vertex = {0,0,0,0,0,255,255,255,255}
	-- function exists mostly as shorthand like the above :D
	local x = x or 0
	local y = y or 0
	local z = z or 0
	local u = u or 0
	local v = v or 0
	local r = r or 255
	local g = g or 255
	local b = b or 255
	local a = a or 255
	return {x,y,z,u,v,r,g,b,a}
end

function getDistance3D(vertex1, vertex2)
	-- 3D distance calculator, expects vertex3D but works with 2D vertices if passed in properly
	-- Using with one argument gets distance from 0,0,0
	local vertex1 = vertex1 or {0,0,0}
	local vertex2 = vertex2 or {0,0,0}

	local arg1 = (vertex2[1]-vertex1[1])^2
	local arg2 = (vertex2[2]-vertex1[2])^2
	local arg3 = (vertex2[3]-vertex1[3])^2

	local dist = math.sqrt(arg1+arg2+arg3)

	return dist
end

function getClockwise(tri2D)
	-- expects a set of three 2D x/y vertices, {{x,y},{x,y},{x,y}}
	-- should be used with tri2D but can be used with any set matching the above format for [1] and [2]
	-- returns true if clockwise, false if ccw (will reverse if this turns out to be wrong)
	-- can also be used with any formatted vertices where x/y are vertex[1]/vertex[2]
	-- won't work right with 3D objects unless testing for clockwise orientation on x/y only
	local vert1 = tri[1]
	local vert2 = tri[2]
	local vert3 = tri[3]
	local clockWiseCalc1 = (vert2[1]-vert1[1])*(vert2[2]+vert1[2])
	local clockWiseCalc2 = (vert3[1]-vert2[1])*(vert3[2]+vert2[2])
	local clockWiseCalc3 = (vert1[1]-vert3[1])*(vert1[2]+vert3[2])
	local clockTotal = clockWiseCalc1 + clockWiseCalc2 + clockWiseCalc3
	if clockTotal >= 0 then
		return true
	else
		return false
	end
end

function getBoxVertices(quad)
	-- usage is : newBox = getBoxVertices(quad)
	-- quad should be set manually, not via table reference, to ensure a fresh init each time
	-- (it's four frigging variables, deal w/ it)
	-- returns a five vertex table, four corners with center
	-- All vertices are {x,y,0} with calculated UVs and RGB set to 255
	-- outputs in vertex3D format
	-- call with no argument for a basic 1x1 quad shape
	-- order of table is {top left, bottom left, center, top right, bottom right}
	local quad = quad or {0,0,1,1}
	local x = quad[1]
	local y = quad[2]
	local w = quad[3]
	local h = quad[4]

	local topLeft = newVertex3D(x,y,0,0,0)
	local bottomLeft = newVertex3D(x,y+h,0,0,1)
	local center = newVertex3D(x+w*0.5,y+h*0.5,0,0.5,0.5)
	local topRight = newVertex3D(x+w,y,0,1,0)
	local bottomRight = newVertex3D(x+w,y+h,0,1,1)
	local basicBox = {topLeft,bottomLeft,center,topRight,bottomRight}
	return basicBox
end

function getBoxTriangles(basicBox) 
	-- runs the above function or uses its formatted output, outputs table with {tri1,tri2,tri3,tri4}
	-- tri format is {vertex1, vertex2, vertex3}
	-- triangles are 1,3,2 then 2,3,5 then 1,4,3 then 4,5,3 to be generated as clockwise
	-- output format for triSet is {tri1,tri2,tri3,tri4}

	local basicBox = basicBox or getBoxVertices()

	local tri3D_1 = {basicBox[1],basicBox[3],basicBox[2]} -- {vertex3D, vertex3D, vertex3D}
	local tri3D_2 = {basicBox[2],basicBox[3],basicBox[5]}
	local tri3D_3 = {basicBox[1],basicBox[4],basicBox[3]}
	local tri3D_4 = {basicBox[4],basicBox[5],basicBox[3]}
	
	local object3D = {tri3D_1,tri3D_2,tri3D_3,tri3D_4}

	return object3D
end

function zSortTriangles(object3D)
	-- accepts and object composed of triangles, returns a set ordered with low z values first, high z values last
	-- triangles are sorted by their minimum, not average, z value
	-- triangles are not internally re-ordered
	local sortedTable = {}
	local pairsTable = {}
	local zValueTable = {}
	
	for i, tri in ipairs(object3D) do
		local averageZ = 0
		for _,vert in ipairs(tri) do
			averageZ = averageZ + vert[3]	
		end
		averageZ = averageZ / 3
		if not pairsTable[averageZ] == nil then
			table.insert(pairsTable[averageZ],tri)
		else
			pairsTable[averageZ] = {}
			table.insert(zValueTable,averageZ)
			table.insert(pairsTable[averageZ],tri)
		end
	end
	--[[for i,set in ipairs(pairsTable) do
		if sortedTable[1] then
			local sorted = false
			if not sorted then 
				table.insert(sortedPairsTable,set)
			end
		else 
			table.insert(sortedPairsTable,set)
		end
	end]]

	table.sort(zValueTable)

	for i, val in ipairs(zValueTable) do
		for i, tri in ipairs(pairsTable[val]) do
			table.insert(sortedTable,tri)
		end
	end

	--[[for i, pair in ipairs(sortedPairsTable) do
		table.insert(sortedTable,pair[2])
	end]]

	return sortedTable
end

function rotate2D(in1,in2,degrees,origin) 
	-- clockwise rotation on arbitrary axis
	-- returns out1, out2 (ie: x,y or x,z etc), assumed usage is : in1, in2 = rotate2D(in1,in2,degrees)
	-- degrees is a number, converted to radian during this function
	-- origin is a number which offsets both axes identically
	-- origin can also be an {a,b} set where a is the origin on the first axis, b is the second
	-- for in1, in2, use x,y for z rot, x,z for y rot, y,z for x rot
	-- use negative degrees for CCW rotation
	-- reversing inputs works for CCW but origin needs to be reversed as well, so this is not ideal

	local degrees = degrees or 90
	local degRad = math.rad(degrees)
	local origin = origin or 0
	
	local off1, off2 = 0,0

	if type(origin) == 'table' then
		off1 = origin[1]
		off2 = origin[2]
	else
		off1, off2 = origin, origin
	end

	local in1 = in1 - off1
	local in2 = in2 - off2

	out1 = in1*math.cos(degRad)-in2*math.sin(degRad)+off1
	out2 = in1*math.sin(degRad)+in2*math.cos(degRad)+off2

	return out1, out2
end

function isometricConversion(vertex3D) 
	-- input is x,y,z, output is x,y,z
	-- calculated z value is based on x/y
	-- rotation and scaling should be done in the original data set, not here
	local x = vertex3D[1] or 0
	local y = vertex3D[2] or 0
	local z = vertex3D[3] or 0

	local xOffsetX = 0.75*x
	local xOffsetY = 0.5*x
	local yOffsetX = -0.75*y
	local yOffsetY = 0.5*y
	local zOffset = 0.75*z

	local xOut = (xOffsetX + yOffsetX)
	local yOut = (xOffsetY + yOffsetY + zOffset)
	local zOut = (x+y)/2

	return xOut, yOut, zOut
end

function iso_getMeshVertices(object) 
	-- Converts an object3D to a vertexDrawSet with isometric converstion
	local isoSet = {}
	local orderedSet = {}
	local vertexDrawSet = {}
	for i, t in ipairs(object) do -- t = triangle
		local tri = {}
		for _,v in ipairs(t) do -- v = vertex
			local outVert = {}
			local x,y,z = isometricConversion(v)
			outVert = newVertex2D(x,y,z,v[4],v[5],v[6],v[7],v[8],v[9])
			table.insert(tri,outVert)
		end
		table.insert(isoSet,tri)
	end
	orderedSet = zSortTriangles(isoSet)
	for i, tri in ipairs(orderedSet) do
		for i, vert in ipairs(tri) do
			table.remove(vert,3)
			table.insert(vertexDrawSet,vert)
		end
	end
	return vertexDrawSet
end

function love.load(arg)
	debug = false
	love.graphics.setDefaultFilter('linear','nearest')

	music = love.audio.newSource('assets/cavern.xm','static')
	music:play()

	vertCount = 0
	triCount = 0
	
	testMapTriangles = {}

	mapSize = 5

	for x=1,mapSize do
		for y=1,mapSize do
			local testBox = getBoxVertices({x-1,y-1,1,1})
			if x == 1 and y == 1 then 
				testMapTriangles = getBoxTriangles(testBox)
			else
				for i, tri in ipairs(getBoxTriangles(testBox)) do
					table.insert(testMapTriangles,tri)
				end
			end
		end
	end

	maxLightDistance = 3
	minLighting = 92
	t_light = {4,2.5,-0.1}

	for i, tri in ipairs(testMapTriangles) do
		vertCount = vertCount + 3
		triCount = triCount + 1
		for j, vert in ipairs(tri) do
			vert[3] = love.math.random()
			local x,y,z = isometricConversion(vert)
			
			local lighting = minLighting
			local lightDistance = getDistance3D(vert,t_light)
			if lightDistance <= maxLightDistance then
				lighting = lighting+(1-lightDistance/maxLightDistance)*(255-minLighting)
			end
			vert[6],vert[7],vert[8] = lighting, lighting, lighting
		end
	end

	testBG = love.graphics.newImage('assets/testBG.png')

	testBoxTexture = love.graphics.newImage('assets/boxTex2.png')
	testBoxMesh = love.graphics.newMesh(vertCount,'triangles')
	testBoxMesh:setVertices(iso_getMeshVertices(testMapTriangles))
	testBoxMesh:setTexture(testBoxTexture)

end

function love.update(dt)
	music:play()

	if started then
		rotateAmount = 0
	else
		rotateAmount = .001
	end

	if love.keyboard.isDown('left') then
		rotateAmount = -1
	elseif love.keyboard.isDown('right') then
		rotateAmount = 1
	end
	
	local editedVertSet = {}
	for i, tri in ipairs(testMapTriangles) do
		for i, vert in ipairs(tri) do
			if not editedVertSet[vert] then
				local newX, newY = rotate2D(vert[1],vert[2],rotateAmount,mapSize/2)
				local x, y, z = isometricConversion(vert)
				vert[1], vert[2] = newX, newY
				editedVertSet[vert] = true
				local lighting = minLighting
				local lightDistance = getDistance3D(vert,t_light)
				if lightDistance <= maxLightDistance then
					lighting = lighting+(1-lightDistance/maxLightDistance)*(255-minLighting)--lighting = (1-lightDistance/maxLightDistance)^5*255
				end
				vert[6],vert[7],vert[8] = lighting, lighting, lighting
			end
		end
	end
	testBoxMesh:setVertices(iso_getMeshVertices(testMapTriangles))

	started = true
end

function love.draw(dt)
	--love.graphics.setColor(96,96,96)
	--love.graphics.rectangle('fill',0,0,love.graphics.getWidth(),love.graphics.getHeight())
	love.graphics.setColor(255,255,255)
	love.graphics.draw(testBoxMesh,love.graphics.getWidth()/2,0,0,love.graphics.getWidth()/10,love.graphics.getWidth()/10)
end